"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var t = __importStar(require("@babel/types"));
var generator_1 = __importDefault(require("@babel/generator"));
var css_1 = require("@compiled/css");
var utils_1 = require("@compiled/utils");
var ast_builders_1 = require("./ast-builders");
var getInterpolation = function (expression, state) {
    if (t.isIdentifier(expression) && state.declarations) {
        var declaration = state.declarations[expression.name];
        if (t.isVariableDeclaration(declaration) && declaration.kind === 'const') {
            // We only want to pick out variable declarations that are constants
            var potentialValue = declaration.declarations[0].init;
            if (t.isStringLiteral(potentialValue) ||
                t.isNumericLiteral(potentialValue) ||
                t.isObjectExpression(potentialValue)) {
                return potentialValue;
            }
        }
    }
    return expression;
};
var normalizeContentValue = function (value) {
    if (value.charAt(0) !== '"' && value.charAt(0) !== "'") {
        return "\"" + value + "\"";
    }
    return value;
};
var extractKey = function (node) {
    if (t.isIdentifier(node)) {
        return node.name;
    }
    if (t.isStringLiteral(node)) {
        return node.value;
    }
    throw new Error('not supported');
};
var extractObjectExpression = function (node, state) {
    var variables = [];
    var css = '';
    node.properties.forEach(function (prop) {
        if (t.isObjectProperty(prop)) {
            // Don't use prop.value directly as it extracts constants from identifiers if needed.
            var propValue = getInterpolation(prop.value, state);
            var key = extractKey(prop.key);
            var value = '';
            if (t.isStringLiteral(propValue)) {
                value = key === 'content' ? normalizeContentValue(propValue.value) : propValue.value;
            }
            else if (t.isNumericLiteral(propValue)) {
                value = css_1.addUnitIfNeeded(key, propValue.value);
            }
            else if (t.isObjectExpression(propValue)) {
                var result = extractObjectExpression(propValue, state);
                css += key + " { " + result.css + " }";
                variables = variables.concat(result.variables);
                return;
            }
            else if (t.isTemplateLiteral(propValue)) {
                // eslint-disable-next-line @typescript-eslint/no-use-before-define
                var result = extractTemplateLiteral(propValue, state);
                value = result.css;
                variables = variables.concat(result.variables);
            }
            else if (t.isExpression(propValue)) {
                var variableName = "--var-" + utils_1.hash(generator_1.default(propValue).code);
                variables.push({ name: variableName, expression: propValue });
                value = "var(" + variableName + ")";
            }
            else {
                throw new Error("Not supported.");
            }
            css += utils_1.kebabCase(key) + ": " + value + ";";
        }
        else if (t.isSpreadElement(prop) && t.isIdentifier(prop.argument)) {
            var declaration = (state.declarations || {})[prop.argument.name];
            if (t.isVariableDeclaration(declaration) &&
                t.isObjectExpression(declaration.declarations[0].init)) {
                // The declaration is in this module - let's use it!
                var declarationValue = declaration.declarations[0].init;
                var result = extractObjectExpression(declarationValue, state);
                css += result.css;
                variables = variables.concat(result.variables);
                return;
            }
            else {
                throw new Error("Declaration for " + prop.argument.name + " could not be found.");
            }
        }
    });
    return { css: css, variables: variables };
};
var extractTemplateLiteral = function (node, state) {
    var variables = [];
    var css = node.quasis.reduce(function (css, q, index) {
        var interpolation = getInterpolation(node.expressions[index], state);
        if (t.isStringLiteral(interpolation) || t.isNumericLiteral(interpolation)) {
            return css + q.value.raw + interpolation.value;
        }
        if (t.isObjectExpression(interpolation)) {
            var result = extractObjectExpression(interpolation, state);
            variables = variables.concat(result.variables);
            return css + result.css;
        }
        if (interpolation) {
            var variableName = "--var-" + utils_1.hash(generator_1.default(interpolation).code);
            var nextQuasis = node.quasis[index + 1];
            var before = css_1.cssBeforeInterpolation(css + q.value.raw);
            var after = css_1.cssAfterInterpolation(nextQuasis.value.raw);
            var expression = before.variablePrefix || after.variableSuffix
                ? // When there is a prefix/suffix we want to ensure the interpolation at least
                    // resolves to an empty string - so we short circuit it to one.
                    t.logicalExpression('||', interpolation, t.stringLiteral(''))
                : interpolation;
            if (before.variablePrefix) {
                expression = ast_builders_1.joinExpressions(t.stringLiteral(before.variablePrefix), expression, null);
            }
            if (after.variableSuffix) {
                expression = ast_builders_1.joinExpressions(expression, t.stringLiteral(after.variableSuffix), null);
            }
            nextQuasis.value.raw = after.css; // Removes any suffixes from the next quasis.
            variables.push({ name: variableName, expression: expression });
            return before.css + ("var(" + variableName + ")");
        }
        return css + q.value.raw;
    }, '');
    return { css: css, variables: variables };
};
exports.buildCss = function (node, state) {
    if (t.isStringLiteral(node)) {
        return { css: node.value, variables: [] };
    }
    if (t.isTemplateLiteral(node)) {
        return extractTemplateLiteral(node, state);
    }
    if (t.isObjectExpression(node)) {
        return extractObjectExpression(node, state);
    }
    throw new Error('Unsupported node.');
};
//# sourceMappingURL=css-builders.js.map